@@ -8,7 +8,7 @@
 
 (require 'mew)
 
-;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
+;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 ;;;
 ;;; Variables
 ;;;
@@ -22,9 +22,9 @@
 (defvar mew-refile-last-folder nil
   "Folder name previously you refiled")
 
-;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
+;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 ;;;
-;;; initialize function
+;;; Initialize function
 ;;;
 
 (defun mew-refile-setup ()
@@ -34,10 +34,10 @@
   ;; load from alist
   (or mew-refile-from-alist
       (setq mew-refile-from-alist (mew-lisp-load mew-refile-from-file)))
-  (add-hook 'kill-emacs-hook (function mew-refile-clean-up)))
+  (add-hook 'kill-emacs-hook 'mew-refile-clean-up))
 
 (defun mew-refile-clean-up ()
-  (remove-hook 'kill-emacs-hook (function mew-refile-clean-up))
+  (remove-hook 'kill-emacs-hook 'mew-refile-clean-up)
   (if (and mew-refile-from-alist
 	   (member 'mew-refile-guess-by-from mew-refile-guess-control))
       (mew-lisp-save mew-refile-from-file mew-refile-from-alist))
@@ -47,7 +47,7 @@
   (setq mew-refile-from-alist nil)
   (setq mew-refile-msgid-alist nil))
 
-;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
+;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 ;;;
 ;;; Guess functions 
 ;;;
@@ -70,29 +70,25 @@
   (let ((funcs mew-refile-guess-control) ret guess info stop)
     (catch 'last
       (while funcs
-	(cond
-	 ;; func is control function
-	 ((string-match "ctrl" (symbol-name (car funcs)))
-	  (when (setq ret (funcall (car funcs) guess auto))
-	    (setq stop t)
-	    (or show-all (throw 'last t))))
-	 ;; func is guess function
-	 (t
-	  (setq ret (funcall (car funcs)))))
-	;; now, ret is the return value from a function.
-	;; make sure ret is a list.
-	(setq ret (or (and (listp ret) ret) (and ret (list ret))))
-	;; shape the guess: (guess1 guess2 ...)
-	;;            info: (info1 info2 ...)
+	(if (string-match "ctrl" (symbol-name (car funcs)))
+	    ;; func is control function
+	    (when (setq ret (funcall (car funcs) guess auto))
+	      (setq stop t)
+	      (or show-all (throw 'last t)))
+	  ;; func is guess function
+	  (setq ret (funcall (car funcs))))
+	(unless (listp ret) (setq ret (list ret)))
 	(setq info (nconc info (list (cons (car funcs) ret))))
-	(unless stop (setq guess (nconc (reverse ret) guess)))
+	(unless stop
+	  (while ret
+	    (mew-addq guess (car ret))
+	    (setq ret (cdr ret))))
 	(setq funcs (cdr funcs))))
     (setq guess (nreverse guess))
-    (if mew-refile-ctrl-multi
-	(cons (mew-uniq-list guess) info)
-      (cons (list (car guess)) info))))
+    (if (not mew-refile-ctrl-multi) (setq guess (list (car guess))))
+    (cons guess info)))
 
-;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
+;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 ;;;
 ;;; Guess control functions
 ;;;
@@ -103,12 +99,57 @@
 (defun mew-refile-ctrl-throw (guess auto)
   (if guess "stop"))
 
-;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
+;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 ;;;
-;;; By alist returns: (guess1 guess2 ...) or nil
+;;; Dispatcher
+;;;
+
+(defun mew-refile-guess-from-dispatch (func &optional addr)
+  "Dispatcher to make mew-refile-guess-by-from-* consider
+mew-refile-guess-from-me-is-special."
+  (let ((from (downcase (or addr (mew-header-parse-address mew-from:) ""))))
+    ;; if From: is my address, addr is the list extracted from To:, Cc:
+    (if (and mew-refile-guess-from-me-is-special
+	     (mew-is-my-address mew-regex-my-address-list from))
+	(let ((addrs (mew-header-parse-address-list mew-refile-guess-key-list))
+	      ret adr)
+	  (while addrs
+	    (setq adr (funcall func (car addrs)))
+	    (mew-addq ret adr)
+	    (setq addrs (cdr addrs)))
+	  (nreverse ret))
+      (funcall func from))))
+
+;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
+;;;
+;;; Guess by folder
+;;;
+
+(defun mew-refile-guess-by-folder ()
+  ;; Guess folders by the To:/Cc: field with folder alist.
+  ;; Mainly used for mailing-list.
+  (let* ((to-cc (mew-header-parse-address-list mew-refile-guess-key-list))
+	 (proto (mew-sinfo-get-proto))
+	 (case (mew-sinfo-get-case))
+	 (alist (mew-proto-folder-alist proto case))
+	 ent ret ml-addr ml-name)
+    (while to-cc
+      (setq ml-addr (mew-addrstr-parse-address (or (car to-cc) "")))
+      (when ml-addr
+	(setq ml-name (mew-addrstr-extract-user ml-addr))
+	(setq ent (or (mew-assoc-case-equal ml-addr alist 1)
+		      (mew-assoc-case-equal ml-name alist 1)))
+	(mew-addq ret (nth 0 ent)))
+      (setq to-cc (cdr to-cc)))
+    (nreverse ret)))
+
+;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
+;;;
+;;; Guess by alist
 ;;;
 
 (defun mew-refile-guess-by-alist ()
+  ;; Guess folders by user-configured alist.
   (mew-refile-guess-by-alist1 mew-refile-guess-alist))
 
 (defun mew-refile-guess-by-alist1 (alist)
@@ -118,9 +159,9 @@
       (setq sublist (cdr (car alist)))
       (cond
        ((eq name t)
-	(setq ret (cons sublist ret)))
+	(mew-addq ret sublist))
        ((eq name nil)
-	(or ret (setq ret (cons sublist ret))))
+	(or ret (mew-addq ret sublist)))
        (t
 	(setq header (mew-header-get-value name))
 	(when header
@@ -136,152 +177,129 @@
 		(setq ent (mew-refile-guess-by-alist1 val)))))
 	    (when ent
 	      (if (listp ent)
-		  (setq ret (nconc ent ret) ent nil)
-		(setq ret (cons ent ret))))
+		  (while ent
+		    (mew-addq ret (car ent))
+		    (setq ent (cdr ent)))
+		(mew-addq ret ent)))
 	    (setq sublist (cdr sublist))))))
       (setq alist (cdr alist)))
-    (mew-uniq-list (nreverse ret))))
+    (nreverse ret)))
 
-(defun mew-refile-guess-by-alist2 (regexp field string)
-  (let (match-strings match-list)
-    (string-match regexp field)
-    (setq match-list (cdr (cdr (match-data))))
-    (while (car match-list)
-      (setq match-strings
-	    (cons (substring field
-			     (car match-list) (car (cdr match-list)))
-		  match-strings))
-      (setq match-list (cdr (cdr match-list))))
-    ;; xxx should use replace-match
-    (while (string-match "\\\\\\([1-9]\\)" string)
-	(setq string
-	      (concat (substring string 0 (match-beginning 0))
-		      (nth (- (length match-strings)
-			      (string-to-int (mew-match 1 string)))
-			   match-strings)
-		      (substring string (match-end 0)))))
-    string))
+(defun mew-refile-guess-by-alist2 (regex string template)
+  ;; regex: "\\([^@]+\\)@\\([^.]+\\)\\.ad\\.jp"
+  ;; string: "admin@iij.ad.jp"
+  ;; template: "+net/\\2/\\1"
+  ;; -> "+net/iij/admin"
+  (let (str strs match repl)
+    (string-match regex string)
+    (setq match (cdr (cdr (match-data))))
+    (while match
+      (setq str (substring string (car match) (car (cdr match))))
+      (setq strs (cons str strs))
+      (setq match (cdr (cdr match))))
+    (setq strs (cons nil (nreverse strs))) ;; cons nil for 0th
+    (while (string-match "\\\\\\([1-9]\\)" template)
+      (setq repl (nth (string-to-int (match-string 1 template)) strs))
+      (setq template (replace-match repl nil nil template)))
+    template))
 
-;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
+;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 ;;;
-;;; By newsgroups returns (guess1 guess2 ...) or nil
-;;;
-
-(defun mew-refile-guess-by-newsgroups ()
-  (let ((newsgroups (mew-addrstr-parse-value-list2 
-		     (mew-header-get-value mew-newsgroups:)))
-	ent ret)
-    (when newsgroups
-      (while newsgroups
-	(setq ent (mew-assoc-case-equal (car newsgroups) mew-folder-alist 1))
-	(if ent (setq ret (cons (nth 0 ent) ret)))
-	(setq newsgroups (cdr newsgroups)))
-      (mew-uniq-list (nreverse ret)))))
-
-;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
-;;;
-;;; By folder returns: (guess1 guess2 ...) or nil
-;;;
-
-(defun mew-refile-guess-by-folder ()
-  (let ((to-cc (mew-header-parse-address-list mew-refile-guess-key-list))
-	ent ret ml-addr ml-name)
-    (while to-cc
-      (setq ml-addr (mew-addrstr-parse-address (or (car to-cc) "")))
-      (when ml-addr
-	(setq ml-name (mew-addrstr-extract-user ml-addr))
-	(setq ent (or (mew-assoc-case-equal ml-addr mew-folder-alist 1)
-		      (mew-assoc-case-equal ml-name mew-folder-alist 1)))
-	(if ent (setq ret (cons (nth 0 ent) ret))))
-      (setq to-cc (cdr to-cc)))
-    (mew-uniq-list (nreverse ret))))
-
-;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
-;;;
-;;; By message-thread returns: guess1 or nil
+;;; Guess by thread
 ;;;
 
 (defun mew-refile-guess-by-thread ()
+  ;; Guess folders by thread with alist created by
+  ;; mew-refile-guess-by-thread-learn.
   (let ((msgid (or (mew-header-get-value mew-references:)
 		   (mew-header-get-value mew-in-reply-to:))))
     ;; search for msgid
     (if (and msgid 
 	     (string-match "\\(<[^ \t\n]*>\\)[^>]*\0" (concat msgid "\0")))
-	(nth 1 (assoc (mew-match 1 msgid) mew-refile-msgid-alist)))))
+	(nth 1 (assoc (match-string 1 msgid) mew-refile-msgid-alist)))))
 
-;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
+;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 ;;;
-;;; By from-{folder,alist} returns: guess1 or (guess1 ...) or nil
+;;; Guess by from folder
 ;;;
 
-(defun mew-refile-guess-from-dispatch (func &optional addr)
-  "Dispatcher to make mew-refile-guess-by-from-* consider
-mew-refile-guess-from-me-is-special."
-  (let ((from (downcase (or addr (mew-header-parse-address mew-from:) ""))))
-    ;; if From: is my address, addr is the list extracted from To:, Cc:
-    (if (and mew-refile-guess-from-me-is-special
-	     (mew-is-my-address mew-regex-my-address-list from))
-	(let ((addrs (mew-header-parse-address-list mew-refile-guess-key-list))
-	      ret adr)
-	  (while addrs
-	    (if (setq adr (funcall func (car addrs)))
-		(setq ret (cons adr ret)))
-	    (setq addrs (cdr addrs)))
-	  (mew-uniq-list (nreverse ret)))
-      (funcall func from))))
-
 (defun mew-refile-guess-by-from-folder (&optional addr)
+  ;; Guess folders by the From: field with folders under +from.
   (mew-refile-guess-from-dispatch 'mew-refile-guess-by-from-folder-body addr))
 
-(defun mew-refile-guess-by-from (&optional addr)
-  (mew-refile-guess-from-dispatch 'mew-refile-guess-by-from-body addr))
-
 (defun mew-refile-guess-by-from-folder-body (&optional addr)
-  (let* ((pfix (cond
-		((string= mew-folders-default-folder "" ) "+")
-		((string= mew-folders-default-folder "+") "+")
-		((eq mew-folders-default-folder nil)      "+")
-		(t (file-name-as-directory mew-folders-default-folder))))
+  (let* ((proto (mew-sinfo-get-proto))
+	 (case (mew-sinfo-get-case))
+	 (list (mew-proto-friend-folder-list proto case))
+	 (sep (if (mew-folder-imapp proto)
+		  (mew-imap-separator case)
+		mew-path-separator))
+	 (qsep (regexp-quote sep))
 	 (from (downcase (or addr (mew-header-parse-address mew-from:) "")))
 	 (user (mew-addrstr-extract-user from))
-	 (pfix-regex (concat "^" (regexp-quote pfix) "\\(.*/\\)?"))
-	 (from-regex (concat pfix-regex (regexp-quote from) "/?$"))
-	 (user-regex (concat pfix-regex (regexp-quote user) "/?$")))
+	 (from-regex (concat qsep (regexp-quote from) "$"))
+	 (user-regex (concat qsep (regexp-quote user) "$")))
     (or
-     (mew-refile-match-in-list from-regex mew-folder-list)
-     (mew-refile-match-in-list user-regex mew-folder-list))))
+     (mew-member-match2 from-regex list)
+     (mew-member-match2 user-regex list))))
+
+;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
+;;;
+;;; Guess by from
+;;;
+
+(defun mew-refile-guess-by-from (&optional addr)
+  ;; Guess folders by the From: field with alist created by
+  ;; mew-refile-guess-by-from-learn.
+  (mew-refile-guess-from-dispatch 'mew-refile-guess-by-from-body addr))
 
 (defun mew-refile-guess-by-from-body (&optional addr)
   (let ((from (downcase (or addr (mew-header-parse-address mew-from:) ""))))
     (cdr (assoc from mew-refile-from-alist))))
 
-;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
+;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 ;;;
-;;; By default returns: guess1
+;;; Guess by newsgroup
+;;;
+
+(defun mew-refile-guess-by-newsgroups ()
+  ;; Guess folders by the Newsgroups field with folder alist.
+  (let* ((newsgroups (mew-addrstr-parse-value-list2 
+		     (mew-header-get-value mew-newsgroups:)))
+	 (proto (mew-sinfo-get-proto))
+	 (case (mew-sinfo-get-case))
+	 (alist (mew-proto-folder-alist proto case))
+	 newsgroup ent ret)
+    (while newsgroups
+      (setq newsgroup (car newsgroups))
+      (setq newsgroups (cdr newsgroups))
+      (setq ent (mew-assoc-case-equal newsgroup alist 1))
+      (mew-addq ret (nth 0 ent)))
+    (nreverse ret)))
+
+;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
+;;;
+;;; Guess by default
 ;;;
 
 (defun mew-refile-guess-by-default (&optional addr)
+  ;; Concat +from and user.
   (mew-refile-guess-from-dispatch 'mew-refile-guess-by-default-body addr))
 
 (defun mew-refile-guess-by-default-body (&optional addr)
-  (let* ((pfix (cond
-		((string= mew-folders-default-folder "" ) "+")
-		((string= mew-folders-default-folder "+") "+")
-		((eq mew-folders-default-folder nil)      "+")
-		(t (file-name-as-directory mew-folders-default-folder))))
-	 (from (downcase (or addr (mew-header-parse-address mew-from:) "")))
-	 (user (mew-addrstr-extract-user from)))
+  (let* ((proto (mew-sinfo-get-proto))
+	 (case (mew-sinfo-get-case))
+	 (fld (mew-proto-friend-folder proto case))
+	 (from (downcase (or addr (mew-header-parse-address mew-from:) ""))))
     (if mew-refile-guess-strip-domainpart
-	(concat pfix user)
-      (concat pfix from))))
+	(setq from (mew-addrstr-extract-user from)))
+    (mew-concat-folder fld from case)))
 
-;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
+;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 ;;;
 ;;; Learning functions
 ;;;
 
-;; dispatcher
-;;
 ;; mew-refile-guess-learn (buf result)
 ;;
 ;; buf is message buffer.
@@ -303,11 +321,36 @@ mew-refile-guess-from-me-is-special."
 	  (mew-refile-guess-by-from-learn chosen info))
       (if (member 'mew-refile-guess-by-thread mew-refile-guess-control)
 	  (mew-refile-guess-by-thread-learn chosen info)))))
-;;
-;; learn from "From:" field
-;;
+
+(defun mew-refile-guess-by-thread-learn (chosen info)
+  ;; Create mew-refile-msgid-alist for mew-refile-guess-by-thread.
+  (let* ((msgid (mew-header-get-value mew-message-id:)) 
+	 (folder (car chosen))
+	 ;; ohter people's honest opinion and my honest opinion.
+	 (oho info)
+	 (mho (cdr (assoc 'mew-refile-guess-by-thread info))))
+    (if (and msgid (string-match "<[^ \n>]*>" msgid))
+	(setq msgid (match-string 0 msgid)))
+    (when (and msgid chosen)
+      ;; if my opninion was right, I learn it.
+      ;; or a folder was not in other people's opinion,
+      ;; I accept it.
+      (catch 'match
+	(while chosen
+	  (if (or (member (car chosen) mho)
+		  (not (catch 'find
+		    (while oho
+		      (and (member (car chosen) (car oho)) (throw 'find t))
+		      (setq oho (cdr oho))))))
+	      (throw 'match (setq folder (car chosen))))
+	  (setq chosen (cdr chosen))))
+      (setq mew-refile-msgid-alist
+	    (cons (list msgid folder "??")
+		  (delq (assoc msgid mew-refile-msgid-alist) ;; delq is right
+			mew-refile-msgid-alist))))))
 
 (defun mew-refile-guess-by-from-learn (chosen info)
+  ;; Create mew-refile-from-alist for mew-refile-guess-by-from.
   (let* ((from (downcase (or (mew-header-parse-address mew-from:) "")))
 	 ;; 'my honest opinion' guessed by mew-refile-guess-by-from.
 	 (mho (nth 1 (assoc 'mew-refile-guess-by-from info)))
@@ -331,7 +374,7 @@ mew-refile-guess-from-me-is-special."
       (catch 'match
 	(while chosen
 	  ;; searching a folder anyone couldn't predict.
-	  (unless (mew-refile-guess-member-p (car chosen) info)
+	  (unless (mew-member* (car chosen) info)
 	    (throw 'match (setq folder (car chosen))))
 	  (setq chosen (cdr chosen))))
       ;; If candidate was found, I memorize it.
@@ -341,69 +384,7 @@ mew-refile-guess-from-me-is-special."
 		    (delq (assoc from mew-refile-from-alist) ;; delq is right
 			  mew-refile-from-alist)))))))
 
-;;
-;; learn from msgid
-;;
-
-(defun mew-refile-guess-by-thread-learn (chosen info)
-  (let* ((msgid (mew-header-get-value mew-message-id:)) 
-	 (folder (car chosen))
-	 ;; ohter people's honest opinion and my honest opinion.
-	 (oho info)
-	 (mho (cdr (assoc 'mew-refile-guess-by-thread info))))
-    (if (and msgid (string-match "<[^ \n>]*>" msgid))
-	(setq msgid (mew-match 0 msgid)))
-    (when (and msgid chosen)
-      ;; if my opninion was right, I learn it.
-      ;; or a folder was not in other people's opinion,
-      ;; I accept it.
-      (catch 'match
-	(while chosen
-	  (if (or (member (car chosen) mho)
-		  (not (catch 'find
-		    (while oho
-		      (and (member (car chosen) (car oho)) (throw 'find t))
-		      (setq oho (cdr oho))))))
-	      (throw 'match (setq folder (car chosen))))
-	  (setq chosen (cdr chosen))))
-      (setq mew-refile-msgid-alist
-	    (cons (list msgid folder "??")
-		  (delq (assoc msgid mew-refile-msgid-alist) ;; delq is right
-			mew-refile-msgid-alist))))))
-
-;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
-;;;
-;;; Backyard functions.
-;;;
-
-(defun mew-refile-guess-subfolder-p (parent child)
-  (string-match 
-   (concat  "^" (regexp-quote (file-name-as-directory parent)))
-   child))
-
-(defun mew-refile-guess-member-p (x lst)
-  "Search X in a nested list LST."
-  (catch 'found
-    (while lst
-      (cond 
-       ((listp (car lst))
-	(if (mew-refile-guess-member-p x (car lst))
-	    (throw 'found t)))
-       (t
-	(if (equal x (car lst))
-	    (throw 'found t))))
-      (setq lst (cdr lst)))))
-
-(defun mew-refile-match-in-list (regex lst)
-  "Grep with REGEX in a list of string LST. (first match)"
-  (let ((case-fold-search t))
-    (catch 'found 
-      (while lst
-	(if (string-match regex (car lst))
-	    (throw 'found (car lst)))
-	(setq lst (cdr lst))))))
-
-;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
+;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 ;;;
 ;;; Summary mode
 ;;;
@@ -422,18 +403,56 @@ mew-refile-guess-from-me-is-special."
   (let ((ref (mew-refile-get src)))
     (when ref (setcar ref dst))))
 
-;; mew-refile-decide-folders returns: ((input1 input2...) 
-;;                                     (guess-func-name1 guess1 guess2...)
-;;                                     (guess-func-name2 guess1 guess2...))
-;; that is, 'car' is a list of user chosen folders.
-;;          'cdr' is a alist of opinions by guess functions.
-;; cdr is needed for learning functions.
-;;
+(defun mew-refile-folder-check (folder &optional force-to-create)
+  "A function to see if FOLDER exists.
+Return t if exists or created. Otherwise, return nil."
+  (when (stringp folder)
+    (cond
+     ((mew-folder-popp  folder) nil)
+     ((mew-folder-nntpp folder) nil)
+     ((mew-folder-virtualp folder) nil)
+     ((mew-folder-imapp folder)
+      (let* ((case (mew-sinfo-get-case))
+	     (mailboxes (mew-imap-folder-alist case)))
+	(if (assoc folder mailboxes)
+	    t
+	  (if (and (y-or-n-p (format "%s does not exist. Create it? " folder))
+		   (mew-imap-folder-insert case folder))
+	      (progn
+		(message "%s will be created" folder)
+		t)))))
+     (t ;; local
+      (let ((absdir (mew-expand-folder folder))  ;; /home/Mail/foo
+	    (create-it force-to-create))
+	(when absdir
+	  (if (file-exists-p absdir)
+	      (if (file-directory-p absdir)
+		  t ;; exists
+		(message "%s is a file" folder)
+		nil) ;; xxx exists but a file
+	    (unless create-it
+	      (if (y-or-n-p (format "%s does not exist. Create it? " folder))
+		  (setq create-it t)))
+	    (if (not create-it)
+		nil ;; not created
+	      (mew-make-directory absdir)
+	      (mew-local-folder-insert folder)
+	      (message "%s has been created" folder)
+	      t)))))))) ;; created
 
 (defun mew-refile-decide-folders (buf msg cur-folders &optional auto exfld)
-  (let (learn-info folders ret cands singlep)
+  ;; This functions returns
+  ;;  ((folder1 folder2...)
+  ;;   (func1 guess11 guess12...)  ;; for learning
+  ;;   (func2 guess12 guess22...))
+  (let ((proto (mew-proto-to-refile (mew-summary-folder-name 'ext)))
+	(case (mew-sinfo-get-case))
+	learn-info folders ret cands singlep lst-lst lst fld)
     (save-excursion
       (set-buffer buf)
+      ;; copy to buffer local variables
+      (mew-sinfo-set-case case)
+      (mew-sinfo-set-proto proto)
       (setq learn-info (mew-refile-guess auto)))
     (if auto
 	;; if auto is set, simply use the guess.
@@ -451,27 +470,47 @@ mew-refile-guess-from-me-is-special."
 	(setq cands (list (nth 0 (car learn-info))))))
       (setq cands (delete nil cands))
       (when exfld
+	;; copying, two folders are necessary
 	(if (or (null cands) (equal (list exfld) cands))
-	    (setq cands (list "+")))
-	(setq cands (cons exfld cands)))
-      (setq cands (mew-uniq-list cands))
+	    (setq cands (list exfld proto))
+	  (setq cands (cons exfld cands))))
+      (setq cands (mew-uniq-list cands)) ;; unavoidable
       (if (and singlep (= (length cands) 1))
-	  (setq folders (mew-input-folders (car cands)))
-	(setq folders (mew-input-folders cands))))
+	  (setq cands (car cands)))
+      (unless cands
+	(setq cands (list proto)))
+      (setq folders (mew-input-folders cands proto)))
     ;; check folder existence.
-    (setq folders (mew-uniq-list (mew-split (mew-join "," folders) ?,)))
-    (while folders
-      (if (mew-folder-check (car folders))
-	  (setq ret (cons (car folders) ret)))
-      (setq folders (cdr folders)))
+    (setq lst-lst (mapcar (lambda (x) (mew-split x ?,)) folders))
+    (while lst-lst
+      (setq lst (car lst-lst))
+      (setq lst-lst (cdr lst-lst))
+      (while lst
+	(setq fld (car lst))
+	(setq lst (cdr lst))
+	(if (and fld (not (member fld ret)) (mew-refile-folder-check fld))
+	    (setq ret (cons fld ret)))))
     (cons (nreverse ret) (cdr learn-info)))) ;; return value
 
+;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
+;;;
+;;; Copy
+;;;
+
 (defun mew-summary-copy ()
+  "Put the refile mark(default is 'o') on this message with
+the current folder as a candidate in addition to guessed folders."
  (interactive)
  (mew-summary-msg-or-part
   (mew-summary-not-in-draft
-   (mew-summary-refile-body
-    nil nil nil nil (mew-summary-folder-name)))))
+   (mew-summary-local-or-imap
+    (mew-summary-refile-body
+     nil nil nil nil (mew-summary-folder-name))))))
+
+;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
+;;;
+;;; Refile (aka move)
+;;;
 
 (defun mew-summary-refile (&optional report)
  "Put the refile mark(default is 'o') on this message. If already
@@ -482,77 +521,91 @@ with '\\[universal-argument]', it displays how the refile rules work in Message
  (interactive "P")
  (mew-summary-msg-or-part
   (mew-summary-not-in-draft
-   (if report (mew-summary-refile-report) (mew-summary-refile-body)))))
+   (mew-summary-local-or-imap
+    (if report (mew-summary-refile-report) (mew-summary-refile-body))))))
 
 (defun mew-summary-refile-body (&optional exp-flds auto no-msg no-mark exfld)
-  (let (msg folders cur-folders mark buf learn-info fld tmp delbuf)
+  (let (msg folder folders cur-folders mark
+	    buf learn-info fld tmp delbuf invalidp)
     (save-excursion
       ;; save the cursor position anyway
       (mew-summary-goto-message)
+      (setq invalidp (mew-summary-msg-invalidp))
       ;; on the message
       (setq fld (mew-summary-folder-name))
-      (setq msg (mew-summary-message-number));; msg is never nil
-      (setq mark (mew-summary-get-mark)));; any mark
-    (if exp-flds
-	(setq folders exp-flds)
-      ;; show message if not displayed
-      (if (or auto (null (mew-sinfo-get-disp-msg)))
-	  (save-excursion
-	    (setq buf (generate-new-buffer mew-buffer-prefix))
-	    (setq delbuf t)
-	    (set-buffer buf)
-	    (mew-erase-buffer)
-	    (mew-insert-message fld msg mew-cs-autoconv
-				mew-header-reasonable-size)
-	    (goto-char (point-min))
-	    (if (and (re-search-forward (concat "^$\\|^" mew-subj:) nil t)
-		     (not (looking-at "^$")))
-		(let ((med (point)))
-		  (forward-line)
-		  (mew-header-goto-next)
-		  (mew-header-decode-region mew-subj: med (point)))))
-	;; need to make a cache or a message buffer.
-	(mew-summary-display nil)
-	;; mew-cache-hit should be first since we want to get 
-	;; information form the top level header.
-	(setq buf (or (mew-cache-hit fld msg) (mew-buffer-message))))
-      (when (and (eq mew-mark-refile mark) (get-buffer fld))
-	(save-excursion
-	  (set-buffer fld)
-	  (setq cur-folders (cdr (mew-refile-get msg)))))
-      (setq learn-info (mew-refile-decide-folders
-			buf msg cur-folders auto exfld))
-      (setq folders (car learn-info)))
-    ;; we must prevent refiling a message to +queue
-    (while folders
-      (unless (mew-folder-queuep (car folders))
-	(setq tmp (cons (car folders) tmp)))
-      (setq folders (cdr folders)))
-    (setq folders (nreverse tmp))
-    (setq folders (delete mew-draft-folder folders))
-    ;; mark refile
-    (unless no-mark
-      (when folders
-	(save-excursion
-	  (mew-summary-goto-message)
-	  (or exp-flds auto (mew-refile-guess-learn buf learn-info))
-	  (when (get-buffer fld)
+      (setq msg (mew-summary-message-number)) ;; msg is never nil
+      (setq mark (mew-summary-get-mark))) ;; any mark
+    (cond
+     ((and mark (> (mew-markdb-level mark) (mew-markdb-level mew-mark-refile)))
+      (or no-msg
+	  (message "Cannot mark here because '%s' is stronger than '%s'"
+		   (mew-markdb-name mark) (mew-markdb-name mew-mark-refile))))
+     (invalidp
+      (or no-msg (message "Cannot mark this invalid message")))
+     (t
+      (if exp-flds
+	  (setq folders exp-flds)
+	;; show message if not displayed
+	(if (or auto (null (mew-sinfo-get-disp-msg)))
 	    (save-excursion
-	      (set-buffer fld)
-	      (mew-refile-reset msg)
-	      (mew-refile-set msg folders)))
-	  (mew-mark-unmark))
-	(mew-mark-put-mark mew-mark-refile no-msg) ;; cursor may move
-	(if (mew-virtual-p)
-	    (mew-summary-mark-in-physical fld msg mew-mark-refile))))
-    (if delbuf (mew-remove-buffer buf))
-    ;; memorize last-folder
-    (setq mew-refile-last-folder folders)
-    (set-buffer-modified-p nil)
-    folders)) ;; return value
+	      (setq buf (generate-new-buffer mew-buffer-prefix))
+	      (setq delbuf t)
+	      (set-buffer buf)
+	      (mew-erase-buffer)
+	      (mew-insert-message fld msg mew-cs-autoconv
+				  mew-header-reasonable-size)
+	      (goto-char (point-min))
+	      (if (and (re-search-forward (concat "^$\\|^" mew-subj:) nil t)
+		       (not (looking-at "^$")))
+		  (let ((med (point)))
+		    (forward-line)
+		    (mew-header-goto-next)
+		    (mew-header-decode-region mew-subj: med (point)))))
+	  ;; need to make a cache or a message buffer.
+	  (mew-summary-display nil)
+	  ;; mew-cache-hit should be first since we want to get 
+	  ;; information form the top level header.
+	  (setq buf (or (mew-cache-hit fld msg) (mew-buffer-message))))
+	(when (and (eq mew-mark-refile mark) (get-buffer fld))
+	  (save-excursion
+	    (set-buffer fld)
+	    (setq cur-folders (cdr (mew-refile-get msg)))))
+	(setq learn-info (mew-refile-decide-folders
+			  buf msg cur-folders auto exfld))
+	(setq folders (car learn-info)))
+      ;; we must prevent refiling a message to +queue
+      (while folders
+	(setq folder (car folders))
+	(setq folders (cdr folders))
+	(unless (mew-folder-queuep folder)
+	  (setq tmp (cons folder tmp))))
+      (setq folders (nreverse tmp))
+      (setq folders (delete mew-draft-folder folders))
+      ;; mark refile
+      (unless no-mark
+	(when folders
+	  (save-excursion
+	    (mew-summary-goto-message)
+	    (or exp-flds auto (mew-refile-guess-learn buf learn-info))
+	    (when (get-buffer fld)
+	      (save-excursion
+		(set-buffer fld)
+		(mew-refile-reset msg)
+		(mew-refile-set msg folders)))
+	    (mew-mark-unmark))
+	  (mew-mark-put-mark mew-mark-refile no-msg) ;; cursor may move
+	  (if (mew-virtual-p)
+	      (mew-summary-mark-in-physical fld msg mew-mark-refile))))
+      (if delbuf (mew-remove-buffer buf))
+      ;; memorize last-folder
+      (setq mew-refile-last-folder folders)
+      (set-buffer-modified-p nil)
+      folders)))) ;; return value
 
 (defun mew-summary-refile-report ()
-  (let ((fld (mew-summary-folder-name))
+  (let ((proto (mew-proto-to-refile (mew-summary-folder-name 'ext)))
+	(case (mew-sinfo-get-case))
+	(fld (mew-summary-folder-name))
 	(msg (mew-summary-message-number2))
 	(win (selected-window))
 	(tmpbuf (generate-new-buffer mew-buffer-prefix))
@@ -566,6 +619,8 @@ with '\\[universal-argument]', it displays how the refile rules work in Message
       (mew-erase-buffer)
       (mew-insert-message fld msg mew-cs-autoconv
 			  mew-header-reasonable-size)
+      (mew-sinfo-set-case case)
+      (mew-sinfo-set-proto proto)
       (setq guess (mew-refile-guess nil t)))
     (mew-remove-buffer tmpbuf)
     (mew-window-configure 'message)
@@ -587,23 +642,37 @@ with '\\[universal-argument]', it displays how the refile rules work in Message
        (while guess
 	 (insert (format "%-32s  " (car (car guess))))
 	 (insert (format "return: %s\n"
-			 (mapconcat (function identity) (cdr (car guess)) ",")))
+			 (mapconcat 'identity (cdr (car guess)) ",")))
 	 (setq guess (cdr guess)))))
     (mew-message-clear-end-of)
     (set-buffer-modified-p nil)
     (select-window win)))
 
+;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
+;;;
+;;; Refile again
+;;;
+
 (defun mew-summary-refile-again ()
   "Put a refile mark on this message according to the previous 
 refile folder."
   (interactive)
   (mew-summary-not-in-draft
-   (save-excursion
-     (mew-summary-goto-message)
-     (message "%s to %s"
-	      (mew-summary-message-number)
-	      (mew-join ", " mew-refile-last-folder)))
-   (mew-summary-refile-body mew-refile-last-folder)))
+   (mew-summary-local-or-imap
+    (let (validp)
+      (save-excursion
+	(mew-summary-goto-message)
+	(setq validp (not (mew-summary-msg-invalidp))))
+      (when validp
+	(message "%s to %s"
+		 (mew-summary-message-number)
+		 (mew-join ", " mew-refile-last-folder))
+	(mew-summary-refile-body mew-refile-last-folder))))))
+
+;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
+;;;
+;;; Auto refile
+;;;
 
 (defun mew-summary-auto-refile (&optional mew-mark-review-only)
   "Refile each message in the folder automatically. If 
@@ -614,61 +683,84 @@ If '\\[universal-argument]' is specified, only messages marked with
   (interactive "P")
   (mew-summary-or-thread
    (mew-summary-not-in-draft
-    (mew-decode-syntax-delete)
-    (let ((after-change-functions nil)
-	  (mew-use-highlight-x-face nil)
-	  (lines (mew-count-lines (point-min) (point-max)))
-	  (line 1) (mark nil) msg)
-      (cond
-       (mew-mark-review-only
-	(setq msg (format "Refile all messages marked with '%c'? "
-			  mew-mark-review)))
-       (mew-refile-auto-refile-skip-any-mark
-	(setq msg (format "Refile all non-marked messages? ")))
-       (t
-	(setq msg (format "Refile all messages except those marked with '%c' and '%c'? " mew-mark-refile mew-mark-delete))))
-      (if (and mew-refile-auto-refile-confirm (not (y-or-n-p msg)))
-	  (message "Not refiled")
-	(message "Auto refiling...")
+    (mew-summary-local-or-imap
+     (mew-decode-syntax-delete)
+     (let ((mew-use-highlight-x-face nil)
+	   (lines (mew-count-lines (point-min) (point-max)))
+	   (line 1) (mark nil) msg)
+       (cond
+	(mew-mark-review-only
+	 (setq msg (format "Refile all messages marked with '%c'? "
+			   mew-mark-review)))
+	(mew-refile-auto-refile-skip-any-mark
+	 (setq msg "Refile all non-marked messages? "))
+	(t
+	 (setq msg "Refile messages including marked with weak marks?")))
+       (if (and mew-refile-auto-refile-confirm (not (y-or-n-p msg)))
+	   (message "Not refiled")
+	 (message "Auto refiling...")
+	 (save-window-excursion
+	   (goto-char (point-min))
+	   (while (not (eobp))
+	     (when (looking-at mew-regex-msg-valid)
+	       (setq mark (mew-summary-get-mark))
+	       (if mew-mark-review-only
+		   (and mark
+			(char-equal mark mew-mark-review)
+			(mew-summary-refile-body nil t 'no-msg))
+		 (or (and mark
+			  (or mew-refile-auto-refile-skip-any-mark
+			      (>= (mew-markdb-level mark)
+				  (mew-markdb-level mew-mark-refile))))
+		     (mew-summary-refile-body nil t 'no-msg)))
+	       (if (= (% (/ (* 100 line) lines) 10) 0)
+		   (message "Auto refiling...%s%%" (/ (* 100 line) lines)))
+	       (setq line (1+ line)))
+	     (forward-line))
+	   (message "Auto refiling...done"))))))))
+
+;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
+;;;
+;;; "mx" extension
+;;;
+
+(defun mew-summary-mark-copy ()
+  "\\<mew-summary-mode-map>
+Put the refile mark onto all messages marked with '*', 
+with the current folder as a candidate in addition to guessed folders.
+This is very convenient to refile all messages picked by '\\[mew-summary-search-mark]'."
+  (interactive)
+  (mew-summary-not-in-draft
+   (mew-summary-local-or-imap
+    (let ((mew-use-highlight-x-face nil)
+	  last)
+      (message "Mark copying...")
+      (save-excursion
 	(save-window-excursion
 	  (goto-char (point-min))
-	  (while (not (eobp))
-	    (when (looking-at mew-regex-msg)
-	      (setq mark (mew-summary-get-mark))
-	      (if mew-mark-review-only 
-		  (and (eq mark mew-mark-review);; xxx char-equal?
-		       (mew-summary-refile-body nil t 'no-msg))
-		(or (eq mark mew-mark-refile)
-		    (eq mark mew-mark-delete)
-		    (and mark mew-refile-auto-refile-skip-any-mark)
-		    (mew-summary-refile-body nil t 'no-msg)))
-	      (if (= (% (/ (* 100 line) lines) 10) 0)
-		  (message "Auto refiling...%s%%"
-			   (/ (* 100 line) lines)))
-	      (setq line (1+ line)))
+	  (while (re-search-forward mew-regex-msg-valid-review nil t)
+	    (setq last (mew-summary-refile-body
+			last nil 'no-msg nil (mew-summary-folder-name)))
 	    (forward-line))
-	  (message "Auto refiling...done")))))))
-
-;;
-;; "mx" extension
-;;
+	  (message "Mark copying...done")))))))
 
 (defun mew-summary-mark-refile ()
   "\\<mew-summary-mode-map>
 Put the refile mark onto all messages marked with '*'.
 This is very convenient to refile all messages picked by '\\[mew-summary-search-mark]'."
   (interactive)
-  (let ((after-change-functions nil)
-	(mew-use-highlight-x-face nil)
-	(last nil))
-    (message "Mark refiling...")
-    (save-excursion
-      (save-window-excursion
-	(goto-char (point-min))
-	(while (re-search-forward mew-regex-msg-review nil t)
-	  (setq last (mew-summary-refile-body last nil 'no-msg))
-	  (forward-line))
-	(message "Mark refiling...done")))))
+  (mew-summary-not-in-draft
+   (mew-summary-local-or-imap
+    (let ((mew-use-highlight-x-face nil)
+	  last)
+      (message "Mark refiling...")
+      (save-excursion
+	(save-window-excursion
+	  (goto-char (point-min))
+	  (while (re-search-forward mew-regex-msg-valid-review nil t)
+	    (setq last (mew-summary-refile-body last nil 'no-msg))
+	    (forward-line))
+	  (message "Mark refiling...done")))))))
 
 (provide 'mew-refile)
 
